/********************************************************************  
  This file is responsible for Parsing the TA graph/tuple files.
  
  The graph/tuple file should only contain the 'FACT TUPLE :'
  section.  However if the '-r' switch is passed to the program
  then the input graph/tuple file should only contain 
  relation tuples ( i.e.   call  P  Q   ) . 
********************************************************************/


#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "common.h"
#include "HashRoutines.h"
#include "globalProcs.h"
#include "globalVars.h"
#include "support.h"
#include "ParseGraph.h"

/***********************************  
  Declarations of static function 
************************************/

static void getTuple (void);
static void processFactTuples ( struct Data *, char * );
static void processRelations ( struct Data * );
static void viewList ( char *, struct ListOfItems * );

/*********************************
   STATIC VARIABLE DEFINITIONS
**********************************/

FILE *fdin;

static char token1[MAX_TOKEN_SIZE];
static char token2[MAX_TOKEN_SIZE];
static char token3[MAX_TOKEN_SIZE];

/*********************************
   STATIC FUNCTION DEFENITIONS
**********************************/


static void getTuple(void)
{
   getToken (token1, fdin );
   getToken (token2, fdin );
   getToken (token3, fdin );
}


static void viewList ( char *name, struct ListOfItems *ptr )
{
   struct ListOfItems *p;
   fprintf ( debugFileNum, "\n List for : %s \n", name );
   fprintf ( debugFileNum, "  --------------------\n");
   
   if ( ptr )
      p = ptr;
   else 
      return;
   
   while ( p )
   {
      fprintf ( debugFileNum, "%s\n", p -> itemPtr -> name );
      p = p -> next;
   }

}

static void processFactTuples ( struct Data *graph, char *inputTupleFile )
{
   char schemeTupleStr[MAX_LINE_SIZE];
   rewind (fdin);
   
   if ( debugMode )
      fprintf ( debugFileNum, "\n$INSTANCE tuples...\n");
      
   while(1)
   {
      if ( feof(fdin) )
      {
         fprintf ( stderr, "ERROR: End of file reached. Try using -r option.\n");
         exit(1);
      }

      getTuple();

      if ( strcmp( token1, INSTANCE) == 0 )
         break;
   }
   
   {
      struct item *itemPtr;
      
      while(1)
      {
         
         if ( debugMode ) {
            fprintf ( debugFileNum, "%-20s %-20s %-20s\n", token1, token2, token3 );
	    fflush(debugFileNum);
	 }
            
         itemPtr = searchHash ( graph, token2 );
         
         if ( ! itemPtr )
         {  
            itemPtr = (struct item *) xmalloc ( sizeof (struct item) );
            itemPtr -> name = (char *) xmalloc ( strlen (token2) + 1 );
            strcpy ( itemPtr -> name, token2 );
            initItem ( itemPtr );
#ifdef LSDIFF
            
            itemPtr -> entityClass = (char *) xmalloc ( strlen (token3) + 1);
            strcpy ( itemPtr -> entityClass , token3 );
            
#endif /* LSDIFF */

            addItemToHashTable ( graph, itemPtr );
         } 
         else
         {
            if ( printParseError )
            {
               fprintf ( stderr, "WARNING: ");
               fprintf ( stderr, "%s is declared twice in file : %s\n", token2, inputTupleFile );
            }   
         }

	 getTuple();

         if ( feof(fdin) || strcmp ( token1, INSTANCE ) != 0 )
            break;

      } /* end while */      
   
   } /* end scope */         
   
} /* end processFactTuples */


static void processRelations ( struct Data *graph )
{
   struct item *itemPtr1;
   struct item *itemPtr2;
   int cnt = 1;
   /*
   %% This logic is here because when the graph file contains the
   %% $INSTANCE info, that info is processed in the 'processFactTuples'
   %% procedure.  Once we break out of that procedure the first token
   %% of the relations has already been read from the file.
   %% If 'processFactTuples' does not get executed then the first token
   %% is not not read in so we read it 'into token1'.
   */
   
   if ( debugMode )
      fprintf ( debugFileNum, "\nRelation tuples...\n");
   
   if ( relationOnlyInput )
      getTuple();
   
   while ( ! feof (fdin) )
   {
       if (  cnt != 1 )
          getTuple();
       
       cnt++;
       
       if ( feof (fdin) )
          break;
       
       if ( debugMode )
          fprintf ( debugFileNum, "%-20s %-20s %-20s\n", token1, token2, token3 );
       
       itemPtr1 = searchHash ( graph, token2 );
       itemPtr2 = searchHash ( graph, token3 );
       
       if ( ! itemPtr1 )
       {
          itemPtr1 = (struct item *) xmalloc ( sizeof (struct item) );
          itemPtr1 -> name = (char *) xmalloc ( strlen(token2) + 1 );
          strcpy ( itemPtr1->name, token2);
          initItem (itemPtr1);
          addItemToHashTable( graph, itemPtr1);
          
          if ( ! relationOnlyInput )
             if ( printParseError )
                fprintf ( stderr, "WARNING: An $INSTANCE is not declared for : %s\n", token2 );
       }
       
       if ( ! itemPtr2 )
       {
          itemPtr2 = (struct item *) xmalloc ( sizeof (struct item) );
          itemPtr2 -> name = (char *) xmalloc ( strlen(token3) + 1 );
          strcpy ( itemPtr2->name, token3);
          initItem (itemPtr2);
          addItemToHashTable( graph, itemPtr2);
          
          if ( ! relationOnlyInput )
             if ( printParseError )
                fprintf ( stderr, "WARNING: An $INSTANCE is not declared for : %s\n", token3 );
       }
       
       
       if ( ! strcmp( token1, CONTAIN) )
       {
          itemPtr1 -> numOfItemsContained ++;
          addToList ( &itemPtr1 -> ptrToConItems, itemPtr2 );
          itemPtr2 -> isContained = TRUE;
          itemPtr2 -> indxIntoParConItemsArray = itemPtr1 -> numOfItemsContained - 1;
          itemPtr2 -> containLevel = itemPtr1 -> containLevel + 1;
       }
       else
       {                     
          itemPtr1 -> numOfRelationsTo ++;
          addToList ( &itemPtr1 -> ptrToRelationsTo, itemPtr2 );
          itemPtr2 -> numOfRelationsFrom ++;
          addToList ( &itemPtr2 -> ptrToRelationsFrom, itemPtr1 );
       } /* end if */
   
   } /* end while */
   
} /* end processRelations */



/*********************************************************
  external function defenitions
**********************************************************/

void Control_ParseGraphFile ( struct Data *graph,
                              FILE *fd,
                              char *inputTupleFile )
{
   fdin = fd;
   
   if ( debugMode )
   {
      fprintf ( debugFileNum, "\nProcessing graph tuples ...\n");
      fprintf ( debugFileNum,   "---------------------------\n");
   }
   
   if ( ! relationOnlyInput )
       processFactTuples ( graph, inputTupleFile );
       
   processRelations ( graph );
}
