/********************************************************************
   This file is responsible for positioning the NEW items/entities.
   
   NEW items/entities are defined as :
       
       - entities that EXIST in the new tuple file but not in the
         old tuple file
         
               OR
               
      - entities that EXIST in both the old tuple file and the
        new tuple file but have different parent containers
        
*********************************************************************/

#include <assert.h>

#include "./../common/common.h"
#include "./../common/globalProcs.h"
#include "./../common/globalVars.h"
#include "./../common/support.h"
#include "PosNewItems.h"

static float windowHeight; /* height of window */
static float windowWidth;  /* width of window */

/***********************************
  Declarations of static function
************************************/

static void initContainerInfo ( struct ContainerInfo * );
static void initialize ( struct item * );
static void cliServInitialize ( struct ContainerInfo * );
static void computeStaticContainerData ( struct ContainerInfo *, struct item * );
static void computeDynamicContainerData ( struct ContainerInfo *, struct item * );
static void cliServComputeStaticContainerData ( struct ContainerInfo *, struct item * );
static void cliServComputeDynamicContainerData ( struct ContainerInfo *, struct item * );
static void positionItemsInContainer ( struct ContainerInfo *, struct item * );
static void positionClients ( struct ContainerInfo *, struct item * ); 
static void positionServers ( struct ContainerInfo *, struct item * ); 
static void position ( struct item *, struct item * );
static void adjustConItemsAttr ( struct item *ptr, float );

/*********************************************************
  static function defenitions
**********************************************************/

static void initContainerInfo ( conPtr )
   struct ContainerInfo *conPtr;
{
   conPtr -> maxXitemsInContainer = 0;
   conPtr -> ptrToMaxXItem = NULL;
   
   conPtr -> containerItemWidth = 0;
   conPtr -> containerItemHeight = 0;
   
   conPtr -> numOfNewItemsInContainer = 0;
   conPtr -> numOfItemsInContainer = 0;
   
   conPtr -> X = 0;

   conPtr -> minServerSpace = 0;
   
   conPtr -> minServerY = 0;
   conPtr -> maxServerY = 0;
   conPtr -> avgServerWidth = 0;
   conPtr -> avgServerHeight = 0;
   
   conPtr -> maxClientY = 0;
   conPtr -> avgClientWidth = 0;
   conPtr -> avgClientHeight = 0;
   
   conPtr -> numOfNewServers = 0;
   conPtr -> numOfServers = 0;
   conPtr -> numOfNewClients = 0;
   conPtr -> numOfClients = 0;
   
   conPtr -> Xclient = 0;
   conPtr -> Xserver = 0;
}

static void initialize ( rootItemPtr )
   struct item *rootItemPtr;
{
   windowHeight = rootItemPtr -> height;
   windowWidth = rootItemPtr -> width;
}

static void cliServInitialize ( container )
   struct ContainerInfo *container;
{
   container -> minServerY = windowHeight;
}


static void computeStaticContainerData ( container, containerPtr )
   struct ContainerInfo *container;
   struct item *containerPtr;
{
   struct ListOfItems *ptr = containerPtr -> ptrToConItems;
   
   container -> containerWidth = containerPtr -> width;
   container -> containerHeight = containerPtr -> height;

   if (debugMode) {
        fprintf ( debugFileNum, "%s contains:\n",
                        containerPtr->name);
        fflush(debugFileNum);
   }

   while (ptr)
   {
      if (debugMode) {
           fprintf ( debugFileNum, "  %s\n", ptr->itemPtr->name);
           fflush(debugFileNum);
      }

      if ( ptr -> itemPtr -> status == newItem ) {
         container -> numOfNewItemsInContainer += 1;
      }
      container -> numOfItemsInContainer += 1;
         
      ptr = ptr -> next;
   }      

   if (debugMode) {
	fprintf ( debugFileNum, "Num items in container '%s': %d - New: %d\n",
			containerPtr->name,
			container -> numOfItemsInContainer,
			container -> numOfNewItemsInContainer);
        fflush(debugFileNum);
   }
}


static void computeDynamicContainerData ( container, containerPtr )
   struct ContainerInfo *container;
   struct item *containerPtr;
{  
   
   if ( container -> ptrToMaxXItem != NULL )
   {
      container -> maxXitemsInContainer = container -> ptrToMaxXItem -> x;
   }
   else  /* should execute this only once during program  */
   {
      struct ListOfItems *list = containerPtr -> ptrToConItems;
      
      container -> maxXitemsInContainer = 0.0;  /* reset value */
      
      while ( list )
      {
         if ( list -> itemPtr -> status == oldItem )
         {
            if ( list -> itemPtr -> x > container -> maxXitemsInContainer )
            {
               container -> maxXitemsInContainer = list -> itemPtr -> x;
               container -> ptrToMaxXItem = list -> itemPtr;
            }
         }
         list = list -> next;
      }
   }  /* end if */
   
   container -> containerItemWidth = 0;
   container -> containerItemHeight = 0;
   
   {
      int cnt = 0;
      struct ListOfItems *list = containerPtr -> ptrToConItems;
      
      while (list)
      {
         if ( list -> itemPtr -> status == oldItem )
         {
            container -> containerItemWidth += list -> itemPtr -> width;
            container -> containerItemHeight += list -> itemPtr -> height;
            cnt += 1;
         }
         list = list -> next;
      }
      
      if ( cnt > 0 )
      {
         container -> containerItemWidth = container -> containerItemWidth /  cnt;
         container -> containerItemHeight = container -> containerItemHeight /  cnt;
      }         
   }  /* end scope */
}

static void cliServComputeStaticContainerData ( container, rootItemPtr )
   struct ContainerInfo *container;
   struct item * rootItemPtr;
{
   struct ListOfItems *list = rootItemPtr -> ptrToConItems;
   
   while (list)
   {
      if ( list -> itemPtr -> kindOfItemInModel == server ) /* it's a server */
      {     
         if ( list -> itemPtr -> status == newItem )
            container -> numOfNewServers += 1;

	 container -> numOfServers += 1;
      }  
      else if ( list -> itemPtr -> kindOfItemInModel == client ) /* it's a client */
      {     
        if ( list -> itemPtr -> status ==  newItem )
           container -> numOfNewClients += 1;

	container -> numOfClients += 1;
                      
      }   
      list = list -> next;
   }

   if (debugMode) {
        fprintf ( debugFileNum, "Num new clients: %d\n",
                        container -> numOfNewClients);
        fprintf ( debugFileNum, "Num new services: %d\n",
                        container -> numOfNewServers);
        fflush(debugFileNum);   
   }
}

static void cliServComputeDynamicContainerData ( container, rootItemPtr )
   struct ContainerInfo *container;
   struct item * rootItemPtr;
{
   int cntCli = 0;
   int cntServ = 0;
   
   struct ListOfItems *list = rootItemPtr -> ptrToConItems;
   
   container -> minServerY = windowHeight;      /* initialization */
   container -> maxServerY = 0.0;               /* initialization */
   container -> minServerSpace = windowHeight;  /* initialization */
   container -> maxClientY = 0.0;               /* initialization */
   
   while (list)
   {
      if ( list -> itemPtr -> kindOfItemInModel == server )
      {
         if ( list -> itemPtr -> status == oldItem )
         {
            container -> avgServerHeight += list -> itemPtr -> height;
            container -> avgServerWidth += list -> itemPtr -> width;
         
            if ( list -> itemPtr -> y < container -> minServerY )
               container -> minServerY = list -> itemPtr -> y;

	    if ( list -> itemPtr -> y + list -> itemPtr -> height >
		container -> maxServerY )
	    {
               container -> maxServerY = list -> itemPtr -> y + 
					 list -> itemPtr -> height;
	    }
            
            if ( ( windowHeight - ( list -> itemPtr -> y + list -> itemPtr -> height ) ) < container -> minServerSpace ) 
               container -> minServerSpace = windowHeight - ( list -> itemPtr -> y + list -> itemPtr -> height );
                    
            cntServ += 1;          
         }  
      }
        
      else if ( list -> itemPtr -> kindOfItemInModel == client ) /* it's a client */
      {    
         if ( list -> itemPtr -> status == oldItem )
         {     
            container -> avgClientHeight += list -> itemPtr -> height;
            container -> avgClientWidth += list -> itemPtr -> width;
              
            if ( list -> itemPtr -> y + list -> itemPtr -> height > container -> maxClientY ) 
               container -> maxClientY = list -> itemPtr -> y + list -> itemPtr -> height;
                      
            cntCli += 1;
              
         }
           
      } /* end if */      
      
      list = list -> next;
   
   } /* end while */
   
   if ( cntServ > 0 ) 
   {
      container -> avgServerHeight = container -> avgServerHeight / cntServ;
      container -> avgServerWidth = container -> avgServerWidth / cntServ;
   }
   else
   {
      container -> avgServerHeight = MAX_CLIENT_OR_SERVER_HEIGHT;
      container -> avgServerWidth = MAX_CLIENT_OR_SERVER_HEIGHT * WIDTH_TO_HEIGHT_RATIO;
   }
     
   if ( cntCli > 0 )
   {
      container -> avgClientHeight = container -> avgClientHeight / cntCli;
      container -> avgClientWidth = container -> avgClientWidth / cntCli;
   }
   else
   {
      container -> avgClientHeight = MAX_CLIENT_OR_SERVER_HEIGHT;
      container -> avgClientWidth = MAX_CLIENT_OR_SERVER_HEIGHT * WIDTH_TO_HEIGHT_RATIO;
   }   
}

#define VISUAL_FACTOR	200 

static float margin, topGap, gap, min_dim;

static void positionItemsInContainer ( container, containerPtr )
   struct ContainerInfo *container;
   struct item *containerPtr;
{
    float boxWidth, boxHeight;
    float xpos, ypos;
    float dx, dy;

    float minWidth = container->containerWidth/12;
    float minHeight = container->containerHeight/12;

    if ( container -> numOfNewItemsInContainer == 0 )
	return;		/* Nothing to do */

    if ( container -> numOfNewItemsInContainer ==
	 container -> numOfItemsInContainer)
    {
	/* Simple horizontal layout over middle */

	boxWidth = (container->containerWidth - margin*2)/
			container->numOfItemsInContainer - gap;
	boxHeight = container->containerHeight/3;

	dx = boxWidth + gap;
	dy = 0.0;

	if (boxWidth < minWidth) {
	    boxWidth = minWidth;
	}

	xpos = gap*2;
	ypos = container->containerHeight * .5;
    }
    else {
	/* Simple vertical layout at right */

	while (1) {
	    boxWidth = container->containerWidth - 
			container->maxXitemsInContainer - margin*2;

	    if (boxWidth >= minWidth)
		break;

	    if ( shrink ( containerPtr, container, min_dim, min_dim ) )
                computeDynamicContainerData( container, containerPtr);
	    else {
		boxWidth = minWidth;
		break;
	    }
	}

	boxHeight = (container->containerHeight - margin*2)/
                        container->numOfNewItemsInContainer - gap;

	dx = 0.0;
	dy = boxHeight + gap;

	if (boxHeight < minHeight) {
	    boxHeight = minHeight;
	}

	xpos = container->containerWidth - boxWidth - gap;
	ypos = gap;
    }

    if (debugMode) {
	fprintf ( debugFileNum, 
		"Container: %s (W: %f, H: %f) - Items: (W: %f, H: %f, dx: %f, dy: %f,  X: %f Y: %f)\n", 
		containerPtr->name, 
		container->containerWidth, container->containerHeight,
		boxHeight, boxWidth, dx, dy, xpos, ypos);
	fflush(debugFileNum);
    }

    assert(container->containerWidth > 0);
    assert(container->containerHeight > 0);

    {
    struct ListOfItems *list = containerPtr -> ptrToConItems;
    struct item *ptr;

    /** assign height and width to the new items **/

    while ( list ) {
        ptr = list -> itemPtr;
        if ( ptr -> status == newItem ) {
            if ( ptr -> ptrToConItems )
                copyAttributeValues ( ptr );

            ptr -> height = boxHeight;
            ptr -> width = boxWidth;
            ptr -> x = xpos;
            ptr -> y = ypos;

    	    if (debugMode) {
        	fprintf ( debugFileNum,
                    "Item: %s (%f, %f)\n", ptr->name, xpos, ypos);
		fflush(debugFileNum);
	    }

            if ( ptr -> ptrToConItems )
                adjustConItemsAttr ( ptr, ptr -> oldHeight / ptr -> height );

	    xpos += dx;
            ypos += dy;
         }
	 list = list -> next;

     } /* end while */
     } /* end scope */
}


static void adjustConItemsAttr ( ptr, factor )
   struct item *ptr;
   float factor;
{
   struct ListOfItems *list = ptr -> ptrToConItems;
   struct item *p;
   
   while ( list )
   {
      p = list -> itemPtr;
      
      p -> x = p -> x / factor;
      p -> y = p -> y / factor;
      p -> height = p -> height / factor;
      p -> width = p -> width / factor;
      
      if ( p -> ptrToConItems )
         adjustConItemsAttr ( p, factor );
         
      list = list -> next;        
   }
}

/*GF*/
static void positionClients ( container, containerPtr ) 
   struct ContainerInfo *container;
   struct item *containerPtr;
{
    float boxWidth, boxHeight;
    float xpos, ypos, dx;

    int fullLayout = 
	(container -> numOfNewClients == container -> numOfClients ||
	relayoutCliServMode);

    if ( container -> numOfNewClients == 0) 
	return;		/* Nothing to do */

    if (fullLayout) {
	boxWidth = containerPtr->width/container->numOfClients - gap;

	dx = boxWidth + gap;

	if (boxWidth < min_dim) {
	   boxWidth = min_dim;
	   boxHeight = min_dim;
	}
	else {
	    boxHeight = boxWidth * HEIGHT_TO_WIDTH_RATIO;

	    if (boxHeight > containerPtr->y - topGap)
	        boxHeight = containerPtr->y - topGap;

	    if (boxHeight < min_dim)
		boxHeight = containerPtr->y - gap;
	}
    }
    else {
        boxHeight = containerPtr->y - container->maxClientY - margin;

        if (boxHeight < min_dim) {
	    boxHeight = min_dim;
        }

        boxWidth = boxHeight * WIDTH_TO_HEIGHT_RATIO;

	dx = boxWidth + gap;
    }

    assert(boxHeight > 0.0);
    assert(boxWidth > 0.0);

    xpos = containerPtr->x;

    if (fullLayout) {
	if (boxHeight + margin*2 > containerPtr->y)
	    ypos = margin;
	else
	    ypos = containerPtr->y - boxHeight - margin*2;
    }
    else {
        ypos = container->maxClientY + margin;
    }
   
    { 
        struct ListOfItems *list = containerPtr -> parent -> ptrToConItems;
        struct item *ptr;
      
        while ( list )  
        {
           ptr = list -> itemPtr;
           if ( ptr -> status == newItem  || fullLayout)
           {
             if ( ptr -> kindOfItemInModel == client )   /** it's a client **/
             {
                if ( ptr -> ptrToConItems )
                   copyAttributeValues ( ptr );
                  
                ptr -> height = boxHeight;
                ptr -> width = boxWidth;
                ptr -> x = xpos;
                ptr -> y = ypos;

		if (debugMode) {
                    fprintf ( debugFileNum,
                        "Client: %s (H:%f, W:%f X:%f Y:%f)\n", 
			ptr->name, boxHeight, boxWidth, xpos, ypos);
                    fflush(debugFileNum);
		}

                if ( ptr -> ptrToConItems )
                   adjustConItemsAttr ( ptr, ptr -> oldHeight / ptr -> height );

                xpos += dx;
             }
           }
           list = list -> next;
        }
        
     } /* end scope */
}

static void positionServers ( container, containerPtr )
   struct ContainerInfo *container;
   struct item *containerPtr;
{
    float boxWidth, boxHeight;
    float xpos, ypos, dx;

    int fullLayout =
        (container -> numOfNewServers == container -> numOfServers ||
        relayoutCliServMode);

    if ( container -> numOfNewServers == 0)
        return;         /* Nothing to do */

    if (fullLayout) {
        boxWidth = containerPtr->width/container->numOfClients - gap;

        dx = boxWidth + gap;

        if (boxWidth < min_dim) {
           boxWidth = min_dim;
           boxHeight = min_dim;
        }
        else {
            boxHeight = boxWidth * HEIGHT_TO_WIDTH_RATIO;

            if (boxHeight < min_dim)
                boxHeight = min_dim;
        }
	ypos = containerPtr->y + containerPtr->height + topGap;
    }
    else {
        boxHeight = container->avgServerHeight;
        boxWidth = container->avgServerWidth;

	dx = boxWidth + margin;

	ypos = container->maxServerY + gap;
    }

    assert(boxHeight > 0.0);
    assert(boxWidth > 0.0);

    xpos = containerPtr->x;

    {
        struct ListOfItems *list = containerPtr -> parent -> ptrToConItems;
        struct item *ptr;

        while ( list )
        {
           ptr = list -> itemPtr;
           if ( ptr -> status == newItem )
           {
             if ( ptr -> kindOfItemInModel == server )   /** it's a server **/
             {
                if ( ptr -> ptrToConItems )
                   copyAttributeValues ( ptr );

                ptr -> height = boxHeight;
                ptr -> width = boxWidth;
                ptr -> x = xpos;
                ptr -> y = ypos;

                if (debugMode) {
                    fprintf ( debugFileNum,
                        "Service: %s (H:%f, W:%f X:%f Y:%f)\n",
                        ptr->name, boxHeight, boxWidth, xpos, ypos);
                    fflush(debugFileNum);
                }

                if ( ptr -> ptrToConItems )
                   adjustConItemsAttr ( ptr, ptr -> oldHeight / ptr -> height );

                xpos += (boxWidth + gap);
             }
           }
           list = list -> next;
        }

     } /* end scope */
}


static void position ( containerPtr, exceptPtr )
   struct item *containerPtr;
   struct item *exceptPtr;
{
   struct ListOfItems *list = containerPtr -> ptrToConItems;
   struct item *ptr;
   
   while ( list )
   {
      ptr = list -> itemPtr;

      if (debugMode) {
           fprintf ( debugFileNum, "Process: %s\n", ptr->name);
           fflush(debugFileNum);
      }
         
      if ( ptr != exceptPtr )
      {
          if ( ptr -> ptrToConItems )
          {
             struct ContainerInfo *container;
             container = (struct ContainerInfo *) xmalloc ( sizeof(struct ContainerInfo) );
	     initContainerInfo ( container );
             computeStaticContainerData ( container, ptr );
             computeDynamicContainerData ( container, ptr );
             positionItemsInContainer ( container, ptr );
             position ( ptr, exceptPtr );
          }
      }   
      list = list -> next;
   }  /* end while */         
}
   
/*********************************************************
  external function defenitions
**********************************************************/

void Control_PosNewItems ( graphPtrs )
   struct SystemPointers *graphPtrs;
{
   struct ContainerInfo *container; 
   
   struct item *rootItemPtr = graphPtrs -> rootItemPtr;
   assert ( rootItemPtr != NULL );
   
   container = (struct ContainerInfo *) xmalloc ( sizeof(struct ContainerInfo) );
   initContainerInfo ( container );
  
   if (debugMode) {
	fprintf ( debugFileNum, "Position new items\n");
	fprintf ( debugFileNum, "------------------\n");
	fflush(debugFileNum);
   } 
   
   if ( layoutKind == nested )
   {
      initialize ( rootItemPtr );
      computeStaticContainerData ( container, rootItemPtr );
      computeDynamicContainerData ( container, rootItemPtr );
      positionItemsInContainer ( container, rootItemPtr );
      position ( rootItemPtr, NULL );
   }
   else   /* layoutKind = cliServ */
   {
      struct item *subSystemPtr = graphPtrs -> subSystemPtr;     
      assert ( subSystemPtr != NULL );
         
      initialize ( rootItemPtr );
      cliServInitialize ( container );
      computeStaticContainerData ( container, subSystemPtr );
      computeDynamicContainerData ( container, subSystemPtr );
      cliServComputeStaticContainerData ( container, rootItemPtr );
      cliServComputeDynamicContainerData ( container, rootItemPtr );

      min_dim = rootItemPtr->width / VISUAL_FACTOR;
      gap = min_dim * 2;
      margin = gap * 2;
      topGap = gap * 3;

      if (debugMode) {
          fprintf ( debugFileNum, 
			"rootWidth:%f min_dim: %f ssy: %f mcy: %f\n",
			rootItemPtr->width, min_dim,
			subSystemPtr->y, container->maxClientY);
          fflush(debugFileNum);
      }

      /* positionItemsInContainer ( container, subSystemPtr ); */
      if (debugMode) {
          fprintf ( debugFileNum, "-- Position Clients --\n");
          fflush(debugFileNum);
      }
      positionClients ( container, subSystemPtr );

      if (debugMode) {
          fprintf ( debugFileNum, "-- Position Services --\n");
          fflush(debugFileNum);
      }
      positionServers ( container, subSystemPtr );

      if (debugMode) {
          fprintf ( debugFileNum, "-- Recurse into SS --\n");
          fflush(debugFileNum);
      }
      position ( subSystemPtr, NULL );
      if (debugMode) {
          fprintf ( debugFileNum, "-- Recurse into Clients and Services --\n");
          fflush(debugFileNum);
      }
      position ( rootItemPtr, subSystemPtr );
   }

   if (debugMode) {
        fprintf ( debugFileNum, "--- END\n\n");
        fflush(debugFileNum);
   }
}
