/*************************************************
   This is the main file for the lsadjust program.  
**************************************************/


#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "./../common/common.h"
#include "./../common/globalVars.h"
#include "./../common/globalProcs.h"
#include "./../common/HashRoutines.h"
#include "./../common/support.h"

#include "./../common/ParseGraph.h"
#include "./../common/ParseLayout.h"
#include "PosNewItems.h"

/********************************************
   macros
*********************************************/

/* name of program is considered an argument to the program */
#define MIN_ARGS 5
#define MAX_ARGS 18   

/********************************************
   Static function declarations
*********************************************/
static void getProgramArgs(int argc, char **argv);
static void displayProgramSyntax (void);
static void writeProgramInfoToLog (void);
static void setup (void);

/*********************************************
   Static variable declarations
**********************************************/
static char *VERSION = "0.4.0";
static char *DATE = "May 1, 1998";
static char *PROGRAM_NAME = "lsadjust";

/*** File Names ****/
static char *oldInGraphFile;
static char *oldInLayoutFile;
static char *newInGraphFile;
static char *newOutLayoutFile;

static char *logFile;
static char *debugFile;

/*** File descriptors ***/
static FILE *oldInGraphNum;
static FILE *oldInLayoutNum;
static FILE *newInGraphNum;
static FILE *newOutLayoutNum;


/*********************************************
   miscellaneous
**********************************************/
static int cont = TRUE;

/*********************************************
   main
**********************************************/

void main( int argc, char **argv )
{

   struct SystemPointers *oldGraphPtrs;
   struct SystemPointers *newGraphPtrs;
   struct Data *oldGraph;
   struct Data *newGraph;

#ifdef DEBUG

   printf ("\n Number of arguments = %d ", argc );
   {
     int i;
     printf ("\nList of arguments");
     printf ("\n-----------------");
     for ( i=0; i < argc; i++ )
         printf ("\n %s", argv[i] );
     printf ("\n"); 
   }
#endif /* DEBUG */

   getProgramArgs (argc, argv);
   setup();
   if ( debugMode )
      writeProgramInfoToLog();

   oldGraphPtrs = (struct SystemPointers *) xmalloc ( sizeof (struct SystemPointers) );
   newGraphPtrs = (struct SystemPointers *) xmalloc ( sizeof (struct SystemPointers) );
   oldGraph = (struct Data *) xmalloc ( sizeof (struct Data) );
   newGraph = (struct Data *) xmalloc ( sizeof (struct Data) );
   
   initHashTable ( oldGraph );

   Control_ParseGraphFile ( oldGraph, oldInGraphNum, oldInGraphFile );

   createAndSetRoot ( oldGraph, oldGraphPtrs );

   if ( layoutKind == cliServ )
     setSubSystemPtr ( oldGraph, oldGraphPtrs, oldInGraphFile,oldSubSysName );

   assignParentInfo ( oldGraphPtrs -> rootItemPtr );
  
   if ( layoutKind == cliServ )
      identifyClientsAndServers ( oldGraphPtrs->rootItemPtr, oldGraphPtrs->subSystemPtr );

   if ( debugMode )
      viewStructure ( oldGraphPtrs );
    
   initHashTable ( newGraph );
   
   Control_ParseLayout ( oldGraph, oldInLayoutNum );

   Control_ParseGraphFile ( newGraph, newInGraphNum, newInGraphFile );

   createAndSetRoot ( newGraph, newGraphPtrs );

   if ( layoutKind == cliServ )
     setSubSystemPtr ( newGraph, newGraphPtrs, newInGraphFile,newSubSysName );

   assignParentInfo ( newGraphPtrs -> rootItemPtr );
  
   if ( layoutKind == cliServ )
      identifyClientsAndServers ( newGraphPtrs->rootItemPtr, newGraphPtrs->subSystemPtr );

   if ( debugMode )
      viewStructure ( newGraphPtrs );

   identifyNewAndOldEntities(oldGraph,oldGraphPtrs,newGraph,newGraphPtrs );

   Control_PosNewItems ( newGraphPtrs );
   
   outputLayout ( newGraphPtrs -> rootItemPtr, newOutLayoutNum ); 

   /** clean up code **/
   
   cleanUp ( oldGraph );
   cleanUp ( newGraph );
   
   if ( oldSubSysName )
      free ( oldSubSysName );
   
   if ( newSubSysName )
      free ( newSubSysName );
      
   fclose ( oldInGraphNum );  
   fclose ( oldInLayoutNum );
   fclose ( newInGraphNum );
   fclose ( newOutLayoutNum );
   
   if ( debugMode )
      fclose ( debugFileNum );
}


static void setup (void)
{ 
   logFile = (char *) xmalloc ( strlen(PROGRAM_NAME) + strlen(".log") + 1 );
   strcpy (logFile, PROGRAM_NAME);
   strcat ( logFile , ".log" );

   debugFile = (char *) xmalloc (strlen(PROGRAM_NAME) + strlen(".debug") + 1 );
   strcpy (debugFile, PROGRAM_NAME); 
   strcat ( debugFile , ".debug" );
   
   oldInGraphNum = fopen ( oldInGraphFile, "r" );
   if ( ! oldInGraphNum )
   {
      printf ("\nCould not open file : %s for read.", oldInGraphFile );
      exit(1);
   }
   
   oldInLayoutNum = fopen ( oldInLayoutFile, "r" );
   if ( ! oldInLayoutNum )
   {
      printf ("\nCould not open file : %s for read.", oldInLayoutFile );
      exit(1);
   }

   newInGraphNum = fopen ( newInGraphFile, "r" );
   if ( ! newInGraphNum )
   {
      printf ("\nCould not open file : %s for read.", newInGraphFile );
      exit(1);
   }

   newOutLayoutNum = fopen ( newOutLayoutFile, "w" );
   if ( ! newOutLayoutNum )
   {
      printf ("\nCould not open file : %s for write.", newOutLayoutFile );
      exit(1);
   }
      
   if ( debugMode == TRUE ) 
   {
      debugFileNum = fopen ( debugFile, "w" );
      if ( !debugFileNum )
      {
         printf ("\nCould not open file : %s for write.", debugFile );
         exit(1);
      }
   }

} /* end setup */


static void getProgramArgs ( int argc, char **argv )
{
   char *argStr;
   int i = 1;

   if ( argc < MIN_ARGS || argc > MAX_ARGS )
   {
      printf ("\nIncorrect number of arguments.");
      displayProgramSyntax();
      cont = FALSE;
      exit(1);
   }
   else
   {
      while (1)
      {

          argStr = argv[i];
          if ( argStr == strchr( argStr, '-' ) )   /* '-' is first char in string
                                                           therefore legal switch     */
          {
             if ( strlen ( argStr ) >= 2 )
             {
                 switch ( argv[i][1] )
                 {
                    case 'H':  if ( strlen ( argStr ) >= 4 )
                               {
                                  switch ( argStr[2] )
                                  {
                                     case 'e':  if ( isReal (argStr+3) )
                                                {
                                                   float tmp = atof ( argStr+3);
                                                   if ( tmp > 0 )
                                                   {
                                                      MIN_CONTAINER_ENTITY_HEIGHT = tmp;
                                                      MIN_CONTAINER_ENTITY_WIDTH = MIN_CONTAINER_ENTITY_HEIGHT * WIDTH_TO_HEIGHT_RATIO;
                                                   }
                                                   else
                                                   {
                                                      printf ("<val> in -H switch must be > 0\n");
                                                      displayProgramSyntax();
                                                      exit(1);
                                                   }
                                                }
                                                else
                                                {
                                                   printf("Invalid value for <val> in -H switch : %s\n", argStr+3 );
                                                   displayProgramSyntax();
                                                   exit(1);
                                                } 
                                                break;
        
                                     case 'c':  if ( isReal ( argStr+3) )
                                                {
                                                   float tmp = atof ( argStr+3 );
                                                   if ( tmp > 0 )
                                                   {
                                                      MIN_CLIENT_HEIGHT = tmp;
                                                      MIN_CLIENT_WIDTH = MIN_CLIENT_HEIGHT * WIDTH_TO_HEIGHT_RATIO;
                                                   }
                                                   else
                                                   {
                                                      printf ("<val> in -H switch must be > 0\n");
                                                      displayProgramSyntax();
                                                      exit(1);
                                                   }
                                                }
                                                else
                                                {
                                                   printf("Invalid value for <val> in -H switch : %s\n", argStr+3 );
                                                   displayProgramSyntax();
                                                   exit(1);
                                                }   
                                                break;
        
                                     case 's':  if ( isReal ( argStr+3) )
                                                {
                                                   float tmp = atof ( argStr+3 );
                                                   if ( tmp > 0 )
                                                   {
                                                      MIN_SERVER_HEIGHT = tmp;
                                                      MIN_SERVER_WIDTH = MIN_SERVER_HEIGHT * WIDTH_TO_HEIGHT_RATIO;
                                                   }
                                                   else
                                                   {
                                                      printf ("<val> in -H switch must be > 0\n" );
                                                      displayProgramSyntax();
                                                      exit(1);
                                                   }
                                                } 
                                                else
                                                {
                                                   printf("Invalid value for <val> in -H switch : %s\n", argStr+3 );
                                                   displayProgramSyntax();
                                                   exit(1);
                                                }
                                                break;


                                     default:   printf ("Incorrect <type> in -H switch : %c", argStr[2] );
                                                displayProgramSyntax();
                                                exit(1); 
                                                break;

                                  }  /* end switch */

                               }
                               else
                               {
                                  printf ("Incorrect use of switch : %s ", argStr );
                                  displayProgramSyntax();
                                  exit(1);
                               }
                               break;
                    
                    case 'S':  if ( isReal ( argStr+2) )
                               {
                                  float tmp = atof ( argStr+2);
                                  if ( tmp > 0 && tmp < 100 )
                                     SHRINK = tmp;
                                  else
                                  {
                                     printf ("Value in -S switch must be between 0 and 100.\n");
                                     displayProgramSyntax();
                                     exit(1);
                                  }
                               }
                               else
                               {
                                  printf ("Invalid value in -S switch : %s .\n", argStr+2);
                                  displayProgramSyntax();
                                  exit(1);
                               }
                               break;
                    
                    case 'c':  if ( strlen (argStr) == 3 )
                               {
                                  switch ( argStr[2] )
                                  {
                                     case '1':  compress = none;
                                                break;
                                     
                                     case '2':  compress = min;
                                                break;
                                     
                                     case '3':  compress = max;
                                                break;
                                                
                                     default:   printf ("Invalid compress type.\n");
                                                displayProgramSyntax();
                                                exit(1);
                                  }
                               }
                               break;
                               
                    case 'e':  printParseError = TRUE;
                               break;
                    
                    case 'm':  if ( strlen (argStr) == 3 )
                               {
                                  switch ( argStr[2] )
                                  {
                                     case '1' :  layoutKind = nested;
                                                 break;
                                     
                                     case '2' :  layoutKind = cliServ;
                                                 break;
                                     
                                     default : printf ("\nInvalid layout model kind selected in -m switch.");
                                               displayProgramSyntax();
                                               exit(1);
                                  
                                  } /* end switch */
                               
                               } 
                               else
                               {
                                  printf ("\nInvalid use of -m switch.");
                                  displayProgramSyntax();
                                  exit(1);
                               }
                               break;

                    case 'r':  relationOnlyInput = TRUE;
                               break;

                    case 'R':  relayoutCliServMode = TRUE;
                               break;
                               
                    case 's':  if ( strlen (argStr) == 3 )
                               {
                                  switch ( argStr[2] )
                                  {
                                     case 'o': i++;
                                               {
                                                 int len = strlen ( argv[i] );
                                                 oldSubSysName = (char *) xmalloc ( sizeof(char) * len + 1 );
                                                 strcpy ( oldSubSysName, argv[i] );
                                               }
                                               break;
                                     
                                     case 'n': i++;
                                               {
                                                 int len = strlen ( argv[i] );
                                                 newSubSysName = (char *) xmalloc ( sizeof(char) * len + 1 );
                                                 strcpy ( newSubSysName, argv[i] );
                                               }
                                               break;
                                     
                                     default : printf ("\nInvalide use of -s switch.");
                                               displayProgramSyntax();
                                               exit(1);
                                               break;
                                  }
                               }
                               else
                               {
                                  printf ("\nInvalide use of -s switch.");
                                  displayProgramSyntax();
                                  exit(1);
                                  break;
                               }
                               break;
                               
                    case 'x':  debugMode = TRUE;
                               break;
                               
                               
    
                    default :  printf ("\nInvalid switch : %s", argStr);
                               displayProgramSyntax();
                               cont = FALSE;
                               exit(1);
    
                 } /* end switch */
             }  /* end if */

          }
          else
          {
          
             break;
          
          }  /* end if */ 

          i++;

      }  /* end while */

      oldInGraphFile = argv[i];
      oldInLayoutFile = argv[i+1];
      newInGraphFile = argv[i+2];
      newOutLayoutFile = argv[i+3];

   }  /* end if */

}  /* end getProgramArgs */


static void displayProgramSyntax ( void )
{
   printf ("\n\n");
   printf ("Program usage:");
   printf ("\n");
   printf ("Version: %s   :  Date: %s", VERSION, DATE );
   printf ("\n\n");
   printf ("%s [options] <old tuple file> <old layout file>", PROGRAM_NAME );
   printf ("\n");
   printf ("                   <new tuple file> <new layout file>");
   printf ("\n\n");
   printf ("options:");
   printf ("\n\n");

   printf ("%-18s %s", "-H<type><val>", "- specify minimum height of <type> to be <val>" );
   printf ("\n");
   printf ("%-18s %s", " ", "- default Height for each <type> is 30" );
   printf ("\n");   
   printf ("%-18s %s", " ", "- description of <type>" );
   printf ("\n");
   printf ("%-18s %s", " ", "  e - entities in container" );
   printf ("\n");
   printf ("%-18s %s", " ", "  c - client entities (if client server model used)" );
   printf ("\n");
   printf ("%-18s %s", " ", "  s - server entities (if client server model used)" );
   printf ("\n\n");

   printf ("%-18s %s", "-S<val>", "- when fitting in new entities shrink by <val>% each iteration");
   printf ("\n");
   printf ("%-18s %s", " ", "  until enough space is created" );
   printf ("\n");
   printf ("%-18s %s", " ", "  Range: 0 < val < 100  , Default value: 1" ); 
   printf ("\n\n");
   
   printf ("%-18s %s", "-c<1 | 2 | 3>", "- specify method to compress static entities in");
   printf ("\n");
   printf ("%-18s %s", " ", "  order to create enough space to fit in new entities." );
   printf ("\n");
   printf ("%-18s %s", " ", "- description of compress types:" );
   printf ("\n");
   printf ("%-18s %s", " ", "  1 - none" );
   printf ("\n");
   printf ("%-18s %s", " ", "  2 - min (relative position and size of all entities is maintained)" );
   printf ("\n");
   printf ("%-18s %s", " ", "  3 - max (relative position is maintained however relative size is not)" );
   printf ("\n");
   printf ("%-18s %s", " ", "- default compress type is 3" );
   printf ("\n\n");
    
   printf ("%-18s %s", "-e", "- output parse warnings" );
   printf ("\n\n");
   
   printf ("%-18s %s", "-m<1 | 2>", "- select a layout model (default is 1)" );
   printf ("\n");
   printf ("%-18s %s", " ", "  1 - nested layout" );
   printf ("\n");
   printf ("%-18s %s", " ", "  2 - client, server, subsystem layout" );
   printf ("\n\n");
   
   printf ("%-18s %s", "-s<o | n> <name>", "- name of subsystem (only for client/server layout model) ");
   printf ("\n");
   printf ("%-18s %s", " ", "  o -  old tuple file" );
   printf ("\n");
   printf ("%-18s %s", " ", "  n -  new tuple file" );
   printf ("\n\n");

   printf ("%-18s %s", "-r", "- <graph files> only contain relation tuples" );
   printf ("\n\n");

   printf ("%-18s %s", "-R", "- Relayout client and servers" );
   printf ("\n\n");

   printf ("%-18s %s", "-x", "- debug mode (creates a file called lslayout.debug)" );
   printf ("\n\n");

} /* end displayProgramSyntax */


static void writeProgramInfoToLog( void )
{
   fprintf ( debugFileNum, "lsadjust Program Environment Information\n");
   fprintf ( debugFileNum, "----------------------------------------\n");
   
   if (oldInGraphNum)
      fprintf ( debugFileNum, "Old graph file : %s \n", oldInGraphFile );
   
   if (oldInLayoutNum)
      fprintf ( debugFileNum, "Old layout file : %s \n", oldInLayoutFile );
   
   if (newInGraphNum)
      fprintf ( debugFileNum, "New graph file : %s \n", newInGraphFile );
   
   if (newOutLayoutNum)
      fprintf ( debugFileNum, "New output layout file : %s \n", newOutLayoutFile );
         
   if (debugFileNum)
      fprintf ( debugFileNum, "Debug file : %s \n", debugFile );
      
   if (relationOnlyInput)
      fprintf ( debugFileNum, "Relation Only mode\n" );
      
   if (debugMode)
      fprintf ( debugFileNum, "DebugMode on\n" );
      
   if ( layoutKind == nested )
      fprintf ( debugFileNum, "Nested Layout Model \n" );
   else
      fprintf ( debugFileNum, "Client Server Layout Model \n" );

   if ( compress == none )
      fprintf ( debugFileNum, "Compress type : none \n" );
   else if ( compress == min )
      fprintf ( debugFileNum, "Compress type : min \n" );
   else
      fprintf ( debugFileNum, "Compress type : max \n" );
      
   fprintf ( debugFileNum, "SHRINK factor : %f \n", SHRINK );

   fprintf ( debugFileNum, "MIN_CONTAINER_ENTITY_HEIGHT = %f \n", MIN_CONTAINER_ENTITY_HEIGHT);
   if ( layoutKind == cliServ )
   {
      fprintf ( debugFileNum, "MIN_CLIENT_HEIGHT = %f \n", MIN_CLIENT_HEIGHT );
      fprintf ( debugFileNum, "MIN_SERVER_HEIGHT = %f \n", MIN_SERVER_HEIGHT );
   }
   
   if ( oldSubSysName )
      fprintf ( debugFileNum, "Name of subsystem in old graph file : %s \n", oldSubSysName );
   
   if ( newSubSysName )   
      fprintf ( debugFileNum, "Name of subsystem in new graph file : %s \n", newSubSysName );

   if ( printParseError )
      fprintf ( debugFileNum, "Print Parse Error mode \n" );
      
} /* end writeProgramInfoToLog */




