/***************************************************************
  This file is responsible for the layout of each level in a 
  container.  This means that these routines assign horizontal 
  positions to each entity within all the levels.  The goal is
  to minimize the length of the edges in the graph layout. 
***************************************************************/


#include <stdio.h>
#include <math.h>

#include "./../common/common.h"
#include "./../common/globalVars.h"
#include "LevelLayout.h"


#define LEVEL_LAYOUT_ITERATIONS 2

/***********************************
  Declarations of static function
************************************/

static void initDesiredPositions ( struct LevelType *, int );
static void assignDesiredPosition ( struct LevelType *, int, int , int, float );
static void computeDesiredPositionsDown ( struct LevelType *, int **, int, int, int );

static void computeDesiredPositionsUp ( struct LevelType *, int **, int, int, int );
static void computeActualPosition ( struct LevelType *, int, int );
static void outputActualPositions ( struct LevelType *, int );
static void outputDesiredPositions ( struct LevelType *, int );


/*********************************************************
  static function defenitions
**********************************************************/

static void outputDesiredPositions ( LevelList, levels )
   struct LevelType *LevelList;
   int levels;
{
   int N,i;
   fprintf ( debugFileNum, "\n\n Output of the initial desired position of the entities in container.\n\n");
   
   for ( N=0; N < levels; N++ )
   {
      fprintf ( debugFileNum, "Level : %d \n", N );
      
      for ( i=0; i < LevelList[N].numOfItemsInLevel; i++ )
      {
          fprintf ( debugFileNum, "   %-20s  %d\n", LevelList[N].list[i] -> name, LevelList[N].list[i] -> desiredPosition );
      }
   } 

}

static void outputActualPositions ( LevelList, levels )
   struct LevelType *LevelList;
   int levels;
{
   int N,i;
   fprintf ( debugFileNum, "\n\n Output of the actual position of the entities in container.\n\n");
   
   for ( N=0; N < levels; N++ )
   {
      fprintf ( debugFileNum, "Level : %d \n", N );
      
      for ( i=0; i < LevelList[N].numOfItemsInLevel; i++ )
      {
          fprintf ( debugFileNum, "   %-20s  %d\n", LevelList[N].list[i] -> name, LevelList[N].list[i] -> actualPosition );
      }
   } 

}



static void initDesiredPositions ( LevelList, levels )
   struct LevelType *LevelList;
   int levels;
{
   int i,j;
   
   for ( i=0; i < levels; i++ )
      for ( j=0; j < LevelList[i].numOfItemsInLevel; j++ )
         LevelList[i].list[j] -> desiredPosition = j + 1;
}


static void assignDesiredPosition ( LevelList, N, i, n, sum )
   struct LevelType *LevelList;
   int N;
   int i;
   int n;
   float sum;
{
   if ( n != 0 )
      if ( LevelList[N].list[i] -> desiredPosition < floor (sum/n) )
         LevelList[N].list[i] -> desiredPosition = floor (sum/n);
}  

static void computeDesiredPositionsUp ( LevelList, adjacencyMatrix, aMdim, curLevel, minLevel )
   struct LevelType *LevelList;
   int **adjacencyMatrix;
   int aMdim;
   int curLevel;
   int minLevel;
{
   int parent, child;
   BOOLEAN edgeExists;
   int n;
   float sum;
   int i,j,k;

   for ( i=0; i < LevelList[curLevel].numOfItemsInLevel; i++ )
   {
      sum = 0.0;
      n = 0; 
      child = LevelList[curLevel].list[i] -> indxIntoParConItemsArray;

      for ( j=curLevel-1; j >= minLevel; j-- )
      {
         for ( k=0; k < LevelList[j].numOfItemsInLevel; k++ )
         {
            parent = LevelList[j].list[k] -> indxIntoParConItemsArray;

            if ( (int) *(adjacencyMatrix + child + aMdim * parent) == 1 )
               edgeExists = TRUE;
            else
               edgeExists = FALSE;

            if ( edgeExists )
            {
               n++;
               sum += LevelList[j].list[k] -> desiredPosition;
            }
         } /* end for */
      } /* end for */
     
      assignDesiredPosition ( LevelList, curLevel, i, n, sum );

   } /* end for */
}


static void computeDesiredPositionsDown ( LevelList, adjacencyMatrix, aMdim, curLevel, totLevels )
   struct LevelType *LevelList;
   int **adjacencyMatrix;
   int aMdim;
   int curLevel;
   int totLevels;
{
   int parent,child;
   BOOLEAN edgeExists;
   int n;
   float sum;
   int i,j,k;
   
   for ( i=0; i < LevelList[curLevel].numOfItemsInLevel; i++ )
   {
      sum = 0.0;
      n = 0;
      parent = LevelList[curLevel].list[i] -> indxIntoParConItemsArray;
      
      for ( j=curLevel+1; j < totLevels; j++ )
      {
         for ( k=0; k < LevelList[j].numOfItemsInLevel; k++ )
         {
            child = LevelList[j].list[k] -> indxIntoParConItemsArray;
            
            if ( (int) *(adjacencyMatrix + child + parent * aMdim) == 1 )
                 edgeExists = TRUE;
            else
                 edgeExists = FALSE;
            
            if ( edgeExists )
            {
               n++;
               sum += LevelList[j].list[k] -> desiredPosition;
            }
                        
         } /* end for */
      
      }  /* end for */
 
      assignDesiredPosition ( LevelList, curLevel, i, n, sum );
   } /* end for */ 
}


static void sortLevelByRelations ( xx, levelNSize )
   struct item *xx[];
   int levelNSize;
{
   int i,j;
   struct item *tmp;
   
   for ( i=0; i < levelNSize-1; i++ )
   {
      for ( j=0; j < levelNSize - i - 1; j++ )
      {
         if ( xx[j] -> indxIntoParConItemsArray > xx[j+1] -> indxIntoParConItemsArray )
         {
            tmp = xx[j];
            xx[j] = xx[j+1];
            xx[j+1] = tmp;
         }
      }   
   }
}

static void computeActualPosition ( LevelList, N, maxItemsInLevel )
   struct LevelType *LevelList;
   int N;
   int maxItemsInLevel;
{
   int levelNSize = LevelList[N].numOfItemsInLevel;
   struct item *position [maxItemsInLevel];
   struct item *xx [levelNSize];
   
   {
      int i;
      for ( i=0; i < maxItemsInLevel; i++ )
         position[i] = NULL;
   }
   
   {
      int i;
      for ( i=0; i < levelNSize; i++ )
         xx[i] = LevelList[N].list[i];
   }
   
   sortLevelByRelations ( xx, levelNSize );
   
   {
      int i;
      int desPos;
      int actPos;
      int dist;

      for ( i=0; i < levelNSize; i++ )
      {
         desPos = xx[i] -> desiredPosition;
         actPos = desPos;
         dist = 1;
         
         while ( position[actPos-1] )
         {
            if ( desPos + dist <= maxItemsInLevel && position[desPos+dist-1] == NULL )
               actPos = desPos + dist;
            else if ( desPos - dist >= 0 && position[desPos-dist-1] == NULL )
               actPos = desPos - dist;

            dist++;
         }

         position [actPos-1] = xx[i];
         
         xx[i] -> actualPosition = actPos;
         xx[i] -> desiredPosition = actPos;
      
      } /* end for */
   } /* end scope */
}

/*********************************************************
  external function defenitions
**********************************************************/

void Control_LevelLayout ( LevelList, adjacencyMatrix, aMdim, levels, maxItemsInLevel )
   struct LevelType *LevelList;
   int **adjacencyMatrix;
   int aMdim;
   int levels;
   int maxItemsInLevel;
{
   int i,L;
   
   initDesiredPositions ( LevelList, levels );
   
   if ( debugMode )
     outputDesiredPositions ( LevelList, levels );

   for ( i=0; i < LEVEL_LAYOUT_ITERATIONS; i++ )
   {
      for ( L=levels-2; L >= 0; L-- )
      {
         computeDesiredPositionsDown ( LevelList, adjacencyMatrix, aMdim, L, levels );
         computeActualPosition ( LevelList, L, maxItemsInLevel );
      }
      
         
      for ( L=1; L < levels; L++ )
      {
         computeDesiredPositionsUp ( LevelList, adjacencyMatrix, aMdim, L, 0 );
         computeActualPosition ( LevelList, L, maxItemsInLevel );
      }
   
   }

   if ( debugMode )
     outputActualPositions ( LevelList, levels );
 
}
 
