/*********************************************************
  This file implements the Sugiygam Barycentric algorithm
  for minimizing the number of edge crossings in a graph.
**********************************************************/


#include <malloc.h>

#include "./../common/common.h"
#include "./../common/globalProcs.h"
#include "SugiyamaLayout.h"

#define SUGIYAMA_ITERATIONS 3


/***********************************  
  Declarations of static function 
************************************/
static void buildDimension ( struct LevelType LevelList[], int **adjacencyMatrix, int aMdim, struct realArray *M[], int N, int NPlus1 );
static void computeAverage ( struct LevelType LevelList[], struct realArray *M[], int N, int NPlus1 );
static void sortDimension ( struct LevelType LevelList[], struct realArray *M[], int N, int NPlus1 ); 
static void swap ( struct LevelType LevelList[], struct realArray *M[], int N, int pos );
static void swapListNodes ( struct LevelType LevelList[], int N, int pos );
static void freeMem ( struct realArray *M[], int size );


/*********************************************************
  static function defenitions
**********************************************************/

static void buildDimension ( struct LevelType LevelList[], int **adjacencyMatrix, int aMdim, struct realArray *M[], int N, int NPlus1 )
{
   int levelNSize = LevelList[N].numOfItemsInLevel;
   int levelNPlus1Size = LevelList[NPlus1].numOfItemsInLevel;
   int i, j;
    
#ifdef DEBUG   
   printf ( "\n\n");
#endif

   for ( i=0; i < levelNSize; i++ )
   {
      int parent = LevelList[N].list[i] -> indxIntoParConItemsArray;
      M[i] = (struct realArray *) xmalloc ( sizeof(struct realArray) );
     
      for ( j=0; j < levelNPlus1Size; j++ )
      {
        int child = LevelList[NPlus1].list[j] -> indxIntoParConItemsArray;
        BOOLEAN edgeExists = FALSE;
        
        if ( (int) *(adjacencyMatrix + parent + child * aMdim ) == 1 || 
             (int) *(adjacencyMatrix + child + parent * aMdim ) == 1  )
        {
           edgeExists = TRUE;
        }
        
        if ( edgeExists )
           M[i] -> list[j] = 1.0;
        else
           M[i] -> list[j] = 0.0;

#ifdef DEBUG           
        printf  (" %f ", M[i]->list[j] );
#endif
      }
#ifdef DEBUG
      printf ( "\n\n");
#endif
   }  
 
}

static void computeAverage ( struct LevelType LevelList[], struct realArray *M[], int N, int NPlus1 )
{
   int levelNSize = LevelList[N].numOfItemsInLevel;
   int levelNPlus1Size = LevelList[NPlus1].numOfItemsInLevel;
   int avg = levelNPlus1Size + 1;
   int i,j;
   float sum;
   int n;
   
   for ( i=0; i < levelNSize; i++ )
   {
      sum = 0.0;
      n = 0;
      
      for ( j=0; j < levelNPlus1Size; j++ )
      {
         if ( M[i] -> list[j] == 1 )
         {
            n++;
            sum = sum + (j + 1);
         }
      }
      
      if ( n == 0 )
         M[i] -> list[avg] = 0.0;
      else
         M[i] -> list[avg] = sum / n;

#ifdef DEBUG
     printf  ("\nM[i] -> list[avg] = %f\n ", M[i]->list[avg] );
#endif
   }
}

static void swapListNodes ( struct LevelType LevelList[], int N, int pos )
{
   struct item *tmpPtr = LevelList[N].list[pos];
   LevelList[N].list[pos] = LevelList[N].list[pos-1];
   LevelList[N].list[pos-1] = tmpPtr;
}

static void swap ( struct LevelType LevelList[], struct realArray *M[], int N, int pos )
{
   struct realArray *tmp = M[pos];
   
   swapListNodes ( LevelList, N, pos );
   
   M[pos] = M[pos-1];
   M[pos-1] = tmp;
}


static void sortDimension ( struct LevelType LevelList[], struct realArray *M[], int N, int NPlus1 ) 
{
   int levelNSize = LevelList[N].numOfItemsInLevel;
   int levelNPlus1Size = LevelList[NPlus1].numOfItemsInLevel;
   int avg = levelNPlus1Size + 1;
   int i,j;

   for ( i=1; i < levelNSize; i++ )
   {
      for ( j=levelNSize-1; j >= i; j-- )
      {
         if ( M[j-1] -> list[avg] > M[j] -> list[avg] )
         { 
#ifdef DEBUG
            printf ("\n Mj-1 avg = %f  Mj avg = %f \n", M[j-1]->list[avg], M[j]->list[avg] );
#endif
            swap ( LevelList, M, N, j );
         }
      }
   }
}

static void freeMem ( struct realArray *M[], int size )
{
    int r;
    for ( r=0; r < size; r++ )
       free ( M[r] );
}

/*********************************************************
  external function defenitions
**********************************************************/

void Control_SugiyamaLayout ( struct LevelType LevelList[], int **adjacencyMatrix, int aMdim, int level )
{

   int i;
   for ( i=0; i < SUGIYAMA_ITERATIONS; i++ )
   {
      int L;
      for ( L=0; L < level - 1; L++ )
      {
         struct realArray  *M[ LevelList[L].numOfItemsInLevel + 2 ];
         
         buildDimension ( LevelList, adjacencyMatrix, aMdim, M, L, L+1 );
         computeAverage ( LevelList, M, L, L+1 );
         sortDimension ( LevelList, M, L, L+1 );
         freeMem ( M, LevelList[L].numOfItemsInLevel );
      }
      
      for ( L=level-1; L > 0; L-- )
      {
         struct realArray  *M[ LevelList[L].numOfItemsInLevel + 2 ];
         
         buildDimension ( LevelList, adjacencyMatrix, aMdim, M, L, L-1 );
         computeAverage ( LevelList, M, L, L-1 );
         sortDimension ( LevelList, M, L, L-1 );
         freeMem ( M, LevelList[L].numOfItemsInLevel );
      }
      for ( L=0; L < level - 1; L++ )
      {
         struct realArray  *M[ LevelList[L].numOfItemsInLevel + 2 ];
         
         buildDimension ( LevelList, adjacencyMatrix, aMdim, M, L, L+1 );
         computeAverage ( LevelList, M, L, L+1 );
         sortDimension ( LevelList, M, L, L+1 );
         freeMem ( M, LevelList[L].numOfItemsInLevel );
      }
      
   }

}
