/*************************************************
   This is the main file for the lslayout program.  
**************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "./../common/ParseGraph.h"
#include "DetermineLayout.h"
#include "./../common/common.h"
#include "./../common/globalVars.h"
#include "./../common/globalProcs.h"
#include "./../common/support.h"
#include "./../common/HashRoutines.h"

/********************************************
   macros
*********************************************/
#define MIN_ARGS 3
#define MAX_ARGS 11

/********************************************
   Static function declarations
*********************************************/
static void getProgramArgs(int argc, char **argv);
static void displayProgramSyntax (void);
static void writeProgramInfoToLog (void);
static void setup (void);
static void emitTAfile ( char *, char * );

/*********************************************
   Static variable declarations
**********************************************/
static char *VERSION = "0.3.4";
static char *DATE = "July 31, 1997";
static char *PROGRAM_NAME = "lslayout";

/*** File Names ****/
static char *inputTupleFile;
static char *outputLayoutFile;

static char *logFile;
static char *debugFile;

/*** File descriptors ***/
static FILE *inputFileNum;
static FILE *outputFileNum;


/*********************************************
   main
**********************************************/

void main( int argc, char **argv )
{
   struct SystemPointers *graphPtrs;
   struct Data *graph;
   
   graphPtrs = (struct SystemPointers *) xmalloc ( sizeof (struct SystemPointers) );
   graph = (struct Data *) xmalloc ( sizeof (struct Data) );
   
   initHashTable ( graph );

   getProgramArgs (argc, argv);
   setup();
   
   if ( debugMode )
     writeProgramInfoToLog();

   Control_ParseGraphFile (graph, inputFileNum, inputTupleFile);
   
   createAndSetRoot ( graph, graphPtrs );  /* creates the $ROOT entity */
   
   
   if ( layoutKind == cliServ )
      setSubSystemPtr ( graph, graphPtrs, inputTupleFile, subsystemName );
   else if ( layoutKind == nested )
      adjustContainLevelAssignments ( graphPtrs -> rootItemPtr -> ptrToConItems, graphPtrs -> rootItemPtr -> containLevel );
      
   if ( debugMode )
      viewStructure ( graphPtrs );
  
   /* this is the call to the driver of the layout routines */ 
   Control_DetermineLayout ( graphPtrs );
  
   /* this outputs the attributes to the layout file */ 
   outputLayout ( graphPtrs -> rootItemPtr, outputFileNum );
  
   if ( emitTA )
      emitTAfile ( inputTupleFile, outputLayoutFile );
   
   /* clean up code */
 
   cleanUp ( graph );
   
   free (subsystemName);
   fclose (inputFileNum);
   fclose (outputFileNum);
   if (debugMode)
      fclose (debugFileNum);

}

/**********************************************************
  If -t switch is passed to the program then this 
  function is called.  This function creates a "ta" file.
***********************************************************/

static void emitTAfile ( inputTupleFile, outputLayoutFile )
   char *inputTupleFile;
   char *outputLayoutFile;
{
   char TAfileName[MAX_TOKEN_SIZE];
   FILE *TAfileNum;
   
   {  
     char *val = (char *) strrchr ( inputTupleFile, '/' );
     char tmp[MAX_TOKEN_SIZE];
     char pref[MAX_TOKEN_SIZE];
     int tmpLen;
     int prefLen;
      
     if ( val )
        strcpy ( tmp, val + 1 );
     else
        strcpy ( tmp, inputTupleFile );
     
     tmpLen = strlen ( tmp );
     
     if ( strchr ( tmp, '.' ) )
        prefLen = (long) strchr ( tmp, '.' ) - (long) tmp;
     else
        prefLen = tmpLen;
        
     strncpy ( pref, tmp, prefLen );
     pref[prefLen] = (char) NULL;
      
     strcpy ( TAfileName, pref);
     strcat ( TAfileName, ".ta");
     
     if ( strcmp ( TAfileName, inputTupleFile ) == 0 )
     {
        strcat ( TAfileName, ".ta" );
        fprintf ( stderr, "The <input tuple file> has a \".ta\" extension.\n  Re-nameing the TA file to %s .\n", TAfileName );
     }
  }
  
  TAfileNum = fopen ( TAfileName, "w" );
  if ( ! TAfileNum )
  {
     printf ("\nCould not open file : %s for write.", TAfileName );
     exit (1);
  }
  else
  {
     fprintf ( TAfileNum, "INCLUDE %s :\n", inputTupleFile );
     fprintf ( TAfileNum, "FACT ATTRIBUTE :\n" );
     fprintf ( TAfileNum, "INCLUDE %s :\n", outputLayoutFile );
  }
  
  fclose ( TAfileNum );
   
}


/********************************************************************
   Setup routine for program.  Opens necessary files that the program
   will use.
*********************************************************************/  

static void setup (void)
{ 
   debugFile = (char *) xmalloc (strlen(PROGRAM_NAME) + strlen(".debug") + 1 );
   strcpy (debugFile, PROGRAM_NAME); 
   strcat ( debugFile , ".debug" );
   
   inputFileNum = fopen ( inputTupleFile, "r" );
   if ( !inputFileNum )
   {
      printf ("\nCould not open file : %s for read.", inputTupleFile );
      exit(1);
   }
   
   outputFileNum = fopen ( outputLayoutFile, "w" );
   if ( !outputFileNum )
   {
      printf ("\nCould not open file : %s for write.", outputLayoutFile );
      exit(1);
   }
   
   if ( debugMode == TRUE ) 
   {
      debugFileNum = fopen ( debugFile, "w" );
      if ( !debugFileNum )
      {
         printf ("\nCould not open file : %s for write.", debugFile );
         exit(1);
      }
   }

} /* end setup */


/**************************************************************
   This routine parses the command line arguments to program.
***************************************************************/

static void getProgramArgs ( int argc, char **argv )
{
   char *argStr;
   int i = 1;

   if ( argc < MIN_ARGS || argc > MAX_ARGS )
   {
      printf ("\nIncorrect number of arguments.");
      displayProgramSyntax();
      exit(1);
   }
   else
   {
      while (1)
      {

          argStr = argv[i];
          if ( argStr == strchr( argStr, '-' ) )   /* '-' is first char in string
                                                           therefore legal switch     */
          {
             if ( strlen ( argStr ) >= 2 )
             {
                 switch ( argv[i][1] )
                 {
                    case 'H':  if ( strlen ( argStr ) > 2 )
                               {
                                  if ( isInt ( argStr + 2 ) )
                                  {
                                     WINDOW_HEIGHT = atoi ( argStr + 2 );
                                  }
                                  else
                                  {
                                     printf ("\nInvalid value in -H switch:  %s", argStr + 2 );
                                     displayProgramSyntax();
                                     exit(1);
                                  }  /* end if */
                               }
                               else
                               {
                                  printf ("\nA value must be given in the -H switch.");
                                  displayProgramSyntax();
                                  exit(1);
                               }  /* end if */
                               break;
                    
                    case 'W':  if ( strlen ( argStr ) > 2 )
                               {
                                  if ( isInt ( argStr + 2 ) )
                                  {
                                     WINDOW_WIDTH = atoi ( argStr + 2 );
                                  }
                                  else
                                  {
                                     printf ("\nInvalid value in -W switch:  %s", argStr + 2 );
                                     displayProgramSyntax();
                                     exit(1);
                                  }  /* end if */
                               }
                               else
                               {
                                  printf ("\nA value must be given in the -W switch.");
                                  displayProgramSyntax();
                                  exit(1);
                               
                               } /* end if */
                               break;
                    
                    case 'e':  printParseError = TRUE;
                               break;
                    
                    case 'm':  if ( strlen (argStr) == 3 )
                               {
                                  switch ( argStr[2] )
                                  {
                                     case '1' :  layoutKind = nested;
                                                 break;
                                     
                                     case '2' :  layoutKind = cliServ;
                                                 break;
                                     
                                     default : printf ("\nInvalid layout model kind selected in -m switch.");
                                               displayProgramSyntax();
                                               exit(1);
                                  
                                  } /* end switch */
                               
                               } 
                               else
                               {
                                  printf ("\nInvalid use of -m switch.");
                                  displayProgramSyntax();
                                  exit(1);
                               }
                               break;

                    case 'r':  relationOnlyInput = TRUE;
                               break;

                    case 's':  i++;
                               {
                                  int len = strlen ( argv[i] );
                                  subsystemName = (char *) xmalloc ( sizeof(char) * len + 1 );
                                  strcpy ( subsystemName, argv[i] );
                               }
                               break;
                               
                    case 't':  emitTA = TRUE;
                               break;
                               
                    case 'x':  debugMode = TRUE;
                               break;
                               
                               
    
                    default :  printf ("\nInvalid switch : %s", argStr);
                               displayProgramSyntax();
                               exit(1);
    
                 } /* end switch */
             }  /* end if */

          }
          else
          {
          
             break;
          
          }  /* end if */ 

          i++;

      }  /* end while */

      inputTupleFile = argv[i];
      outputLayoutFile = argv[i+1];

   }  /* end if */

}  /* end getProgramArgs */


/********************************************************
   Routine to display the program syntax/usage.
*********************************************************/

static void displayProgramSyntax ( void )
{
   printf ("\n\n");
   printf ("Program usage:");
   printf ("\n");
   printf ("Version: %s   Date: %s", VERSION, DATE );
   printf ("\n\n");
   printf ("%s [options] <input tuple file> <output layout file>", PROGRAM_NAME );
   printf ("\n\n");
   printf ("options:");
   printf ("\n\n");
   
   printf ("%-15s %s", "-H<size>", "- specify height of layout window ( default is 600)" );
   printf ("\n\n");
   
   printf ("%-15s %s", "-W<size>", "- specify width of layout window ( default is 800 )" ); 
   printf ("\n\n");
   
   printf ("%-15s %s", "-e", "- output parse warnings" );
   printf ("\n\n");
   
   printf ("%-15s %s", "-m<1 | 2>", "- select a layout model (default is 1)" );
   printf ("\n");
   printf ("%-15s %s", " ", "  1 -  nested layout" );
   printf ("\n");
   printf ("%-15s %s", " ", "  2 -  client, server, subsystem layout" );
   printf ("\n\n");
   
   printf ("%-15s %s", "-r", "- <input graph file> only contains relation tuples" );
   printf ("\n\n");
   
   printf ("%-15s %s", "-s <name>", "- name of subsystem (only for client/server layout model)");
   printf ("\n\n");
   
   printf ("%-15s %s", "-t", "- emit ta file called:  prefix of <input tuple file> + \".ta\" ");
   printf ("\n\n");
   
   printf ("%-15s %s", "-x", "- debug mode (creates a file called lslayout.debug)" );
   printf ("\n\n");
   

} /* end displayProgramSyntax */

static void writeProgramInfoToLog( void )
{
   fprintf ( debugFileNum, "lslayout Program Environment Information\n");
   fprintf ( debugFileNum, "----------------------------------------\n");
   
   if (inputFileNum)
      fprintf ( debugFileNum, "Input tuple file : %s \n", inputTupleFile );
   
   if (outputFileNum)
      fprintf ( debugFileNum, "Output layout file : %s \n", outputLayoutFile );
   
   if (debugFileNum)
      fprintf ( debugFileNum, "Debug file : %s \n", debugFile );
   
   if (subsystemName)
      fprintf ( debugFileNum, "Name of subsystem to passed to program : %s\n", subsystemName );
     
   if (relationOnlyInput)
      fprintf ( debugFileNum, "Relation Only mode\n" );
      
   if (debugMode)
      fprintf ( debugFileNum, "DebugMode on\n" );
      
   if ( layoutKind == nested )
      fprintf ( debugFileNum, "Nested Layout Model \n" );
   else
      fprintf ( debugFileNum, "Client Server Layout Model \n" );
   
   fprintf ( debugFileNum, "layout window height = %i \n", WINDOW_HEIGHT );
   fprintf ( debugFileNum, "layout window width = %i \n", WINDOW_WIDTH );


   if ( printParseError )
      fprintf ( debugFileNum, "Print Parse Error mode \n" );

} /* end writeProgramInfoToLog */

